from .arguments import Arguments, Flag, Flags
from .context import Context, DistroCode, OSCode
from .helpers import ShellHelpers
from .logging import Logger, LogLevel
from .registry import Registry
from .version import Version, VersionPair
from .step import StepExecutor, StepResult, StepResultCategory, StepResultReason
from .step import PackageManager, ShellManager
from .step import StepExecutorExceptionBase, StepExecutorError, StepExecutorWarning, Exceptions
__all__ = ['arguments', 'context', 'helpers', 'logging', 'step', 'version', 'registry']
#Registry is Windows-only.
try:
	import registry
	__all__.append('registry')
except ImportError:
	pass



from .stepbase import StepExecutor
from .stepresult import StepResult, StepResultCategory, StepResultReason
from .shellmanager import ShellManager
from .packagemanager import PackageManager
from .exceptions import StepExecutorExceptionBase, StepExecutorError, StepExecutorWarning, Exceptions
__all__ = ['stepbase', 'shellmanager', 'packagemanager', 'exceptions']



'''Operations on the Windows registry.
'''

#Don't define any of this if we're not running under Windows.
try:
	from .version import Version
	try:
		import winreg as _winreg
	except ImportError:
		import _winreg

	###
	# WINDOWS REGISTRY OPERATIONS
	###
	class Registry(object):
		'''Allows operations on the Windows registry.
		On a non-Windows platform this class' methods do nothing.
		'''

		@classmethod
		def openPath(cls, hKey, keyPath):
			'''Opens the specified key if possible.
			'''

			try:
				return _winreg.OpenKey(hKey, keyPath, 0, _winreg.KEY_READ)
			except WindowsError:
				return None

		@classmethod
		def checkKeyExists(cls, key32, key64, minVersion):
			'''Determines if a subkey >= minVersion exists
			in the given key. minVersion must be a Version object.
			key32 is used first, then key64 if key32 couldn't be found.
			Returns:
				* True if a viable subkey was found.
				* False if key32 and key64 don't exist, or
				no viable subkey could be found.
			'''

			hKey = _winreg.HKEY_LOCAL_MACHINE
			#See if the 32 bit key exists.
			key = Registry.openPath(hKey, key32)
			#Else, see if the 64 bit key exists.
			if key is None:
				key = Registry.openPath(hKey, key64)
			#Otherwise we abort here.
			if key is None:
				return False

			subkeys = []
			#Now enumerate the subkeys...
			subIdx = 0
			while True:
				try:
					subkeys.append(_winreg.EnumKey(key, subIdx))
					subIdx += 1
				except WindowsError:
					break

			#Search for one equal to or over the minimum version.
			for subkeyPath in subkeys:
				try:
					subkeyVersion = Version(subkeyPath)
					if subkeyVersion >= minVersion:
						return True
				except Exception:
					pass
			return False

except ImportError:
	class Registry(object):
		'''Dummy implementation of Registry for
		non-Windows platforms. All methods do nothing.
		'''
		@classmethod
		def openPath(cls, *_):
			'''Does nothing.
			'''
			return None

		@classmethod
		def checkKeyExists(cls, *_):
			'''Does nothing.
			'''
			return False



'''Provide shell-related convenience functions.
'''
import os
import datetime
from ...setup import Context

kDefaultAppName = "basis"

class ShellHelpers(object):
	@classmethod
	def absolutePath(cls, pathString):
		'''Returns the absolute path of the given
		path string, if possible.
		'''
		return os.path.realpath(os.path.expandvars(os.path.expanduser(pathString)))

	@classmethod
	def fileSystemFriendlyDate(cls):
		'''Returns a YYYY-MM-DD date
		that is valid for the common filesystems in use.
		'''
		return datetime.datetime.now().strftime("%Y-%m-%d")

	@classmethod
	def defaultLogFileName(cls, appName=""):
		'''Returns a YYYY-MM-DD date
		that is valid for the common filesystems in use.
		'''
		finalAppName = appName if appName else kDefaultAppName
		return "{0}_log_{1}.log".format(finalAppName, cls.fileSystemFriendlyDate())

	@classmethod
	def loggingDirectory(cls, ctx, appName=""):
		'''Returns logging directory
		as normalized path.
		'''
		assert isinstance(ctx, Context)
		#Depends on the platform.
		finalAppName = appName if appName else kDefaultAppName
		logPath = ""
		if ctx.osIsWindows():
			logPath = "%appdata%\\{0}\\logs"
		elif ctx.osIsMacOS():
			logPath = "~/Library/Logs/{0}"
		elif ctx.osIsLinux():
			logPath = "~/var/log/{0}"

		return cls.absolutePath(logPath.format(finalAppName))

	@classmethod
	def defaultLogFilePath(cls, ctx, appName=""):
		finalAppName = appName if appName else kDefaultAppName
		return cls.loggingDirectory(ctx, finalAppName) + os.path.sep + cls.defaultLogFileName(finalAppName)



'''Unit tests for arguments module.
'''

import unittest
from ...setup.arguments import Arguments, Flag, Flags

class Values:
	'''Possible values for a flag.
	'''
	Disable = 0
	Long = 1
	Short = 2
	NumValues = 3

class ArgumentsTest(unittest.TestCase):
	'''Unit tests for arguments module.
	'''

	def _flagValuesToFlagList(self, allFlags, flagValues):
		'''Turns a list of flag visibility values
		into the actual argument list it represents.
		'''
		result = []
		for i in range(0, len(allFlags)):
			value = flagValues[i]
			flag = allFlags[i]
			if value == Values.Long:
				assert flag.value is not None
				result.append(flag.value)
			elif value == Values.Short and flag.shortValue is not None:
				result.append(flag.shortValue)
		return result

	def _parsedValuesMatchFlagValues(self, allFlags, parsedValues, flagValues):
		result = True
		#For each flag:
		for i in range(0, len(allFlags)):
			flag = allFlags[i]
			value = flagValues[i]
			flagSet = value == Values.Long or (value == Values.Short and flag.shortValue is not None)
			#Test: the parsed values set has a field for this flag.
			if(hasattr(parsedValues, flag.name)):
				#Test: if its flag value is set to be visible,
				#	its parsed value is True. Otherwise, the parsed value
				#	is False.
				flagValue = getattr(parsedValues, flag.name)
				if flagValue != flagSet:
					print("Expected flag {0} to have value {1}, got {2}".format(flag.value, flagSet, flagValue))
					result = False
			else:
				print("Parser is missing expected flag {0}".format(flag.value))
				result = False
		return result

	def _validateParseStep(self, allFlags, flagValues, flagIdx, failedFlags):
		if flagIdx >= len(flagValues):
			#Try this combination.
			#Test: combination does not throw.
			combinationThrew = False
			combinationDoesntMatch = False
			args = self._flagValuesToFlagList(allFlags, flagValues)
			try:
				#Test: combination returns expected settings.
				parsedValues = Arguments.parse(args)
				if not self._parsedValuesMatchFlagValues(allFlags, parsedValues, flagValues):
					print("Argument list {0} didn't match expected values".format(args))
					combinationDoesntMatch = True
			except Exception as e:
				print("Argument list {0} threw exception: {1}".format(args, e))
				combinationThrew = True
			#If either happens, add to the failed list.
			if combinationDoesntMatch or combinationThrew:
				failedFlags.append(tuple(flagValues))
		else:
			#Try each value for this flag and recurse.
			for value in range(0, Values.NumValues):
				flagValues[flagIdx] = value
				self._validateParseStep(allFlags, flagValues, flagIdx+1, failedFlags)

	def _validateParse(self, allFlags, flagValues):
		'''Tries all combinations of flags
		and returns any combinations that failed to validate
		as a list of tuples of Values.
		'''
		failedFlags = []
		self._validateParseStep(allFlags, flagValues, 0, failedFlags)
		return failedFlags

	def testParse(self):
		'''Tests that parser works as expected.
		'''

		#Test: try all flags in every order.
		#For a given combination:
		#	* Combination returns expected settings.
		#	* Combination does not throw.
		flagsValidated = False
		allFlags = Flags.AllBoolean
		flagValues = [Values.Disable] * len(allFlags)
		failedFlags = self._validateParse(allFlags, flagValues)
		flagsValidated = len(failedFlags) < 1

		#Test: Make sure invalid combinations raise an exception.
		invalidFlagsValidated = False
		try:
			invalidFlags = ["--ofsjis", "-0", "swfk28"]
			#TODO: Supress parser errors if verbose test is disabled?
			Arguments.parse(invalidFlags)
			print("Failure: Arguments.parse() didn't raise an exception on invalid input")
		#Oddly enough, this fails if an exception type is specified.
		#pylint: disable=W0702
		except:
			invalidFlagsValidated = True

		self.assertTrue(flagsValidated and invalidFlagsValidated)

	def testExtraFlags(self):
		'''Tests that the extra flags option works.
		'''

		extraFlagName = "extra-flag"
		extraFlag = Flag("Test extra flag.", extraFlagName)
		#Make sure it's set to true when passed in the parse.
		parsedArgs = Arguments.parse([extraFlag.value], (extraFlag,))
		self.assertTrue(parsedArgs.extra_flag,
			"Extra flag '{0}' should be set but isn't".format(extraFlag.value))

if __name__ == "__main__":
	suite = unittest.TestLoader().loadTestsFromTestCase(ArgumentsTest)
	unittest.TextTestRunner(verbosity=2).run(suite)



'''Unit tests for stepbase module.
'''

import unittest
from ...sampleteststeps import TestSteps, AnomalousSteps
from .shellmanager import ShellManagerTest
from .packagemanager import PackageManagerTest
from ....setup import Context, Logger
from ....setup import StepExecutor, StepResult, StepResultReason, StepResultCategory
from ....setup import StepExecutorWarning

#These steps should return True for their install step.
SuccessSteps = ((TestSteps.Pass(), TestSteps.Installed(),
 TestSteps.NotDone(), TestSteps.WrongVersion()), StepResultCategory.Success)
#These steps should not run their install step.
SkipSteps = ((TestSteps.Unnecessary(), TestSteps.AlreadyDone()), StepResultCategory.Skip)
FailSteps = ((TestSteps.Fail(),), StepResultCategory.Fail)
AllSteps = (SuccessSteps, SkipSteps, FailSteps)

SuccessCheckSteps = ((TestSteps.Installed(),), StepResultCategory.Success)
#These steps should not run their install step.
SkipCheckSteps = ((TestSteps.Unnecessary(), TestSteps.AlreadyDone()), StepResultCategory.Skip)
FailCheckSteps = ((TestSteps.Pass(), TestSteps.Fail(), TestSteps.NotDone(), TestSteps.WrongVersion()), StepResultCategory.Fail)
AllCheckSteps = (SuccessCheckSteps, SkipCheckSteps, FailCheckSteps)

class StepExecutorTest(unittest.TestCase):

	def _validateVerb(self, stepExecutor, stepSetAndCategory, performMethodDescription, performMethod):
		'''Checks that the given method of StepExecutor
		performs under unit tests.
		Parmeters:
			* stepExecutor: The StepExecutor being tested.
			* check_only: Boolean. If True this runs unit tests
			as if context.check_only was set.
			* stepSetAndCategory: ((TestStep,), StepResultCategory). Lists the steps to be run and the expected category their results should fall under.
			If a step does not give a result in the expected category, that is a test failure.
			* performMethodDescription: A human-readable name for the method being run.
			* performMethod: The StepExecutor method to be tested.
		Returns:
			* True if performMethod passed all tests, False otherwise.
		Raises:
			* RuntimeError if performMethodDescription isn't set.
		'''
		assert isinstance(stepExecutor, StepExecutor)
		stepSet = stepSetAndCategory[0]
		expectedResultCategory = stepSetAndCategory[1]
		self.assertTrue(expectedResultCategory >= 0 and expectedResultCategory < StepResultCategory.Count,
			"Expected result category '{0}' not a valid StepResultCategory value".format(expectedResultCategory))
		print("")
		if not performMethodDescription:
			raise RuntimeError("_validateVerb(): performMethodDescription parameter can't be empty")
		#Test: Checkonly mode always returns True, even if the check failed.
		validated = True
		for step in stepSet:
			result = performMethod(step)
			if result.category != expectedResultCategory:
				print("Step {0} should return result in category {1} during a {2}, returned {3}".format(step.Name,
						StepResultCategory.Descriptions[expectedResultCategory],
						performMethodDescription,
						result))
				validated = False

		originalPermitWarnings = stepExecutor.context.permit_warnings
		#Also check that nameless steps throw...
		stepExecutor.context.permit_warnings = False
		try:
			performMethod(AnomalousSteps.Nameless)
			print("Method {0} was supposed to raise for nameless steps but didn't".format(performMethodDescription))
			validated = False
		except StepExecutorWarning:
			pass

		#But not when permit-warnings is set.
		stepExecutor.context.permit_warnings = True
		try:
			performMethod(AnomalousSteps.Nameless)
		except StepExecutorWarning:
			print("Method {0} wasn't supposed to raise for nameless steps when context.permit_warnings is True but did anyway".format(performMethodDescription))
			validated = False

		return validated

	def _validateCheck(self, check_only, stepSetAndCategory):
		logger = Logger()
		ctx = Context(logger)
		stepExecutor = StepExecutor(ctx, logger)
		ctx.check_only = check_only
		return self._validateVerb(stepExecutor, stepSetAndCategory, "check", stepExecutor.check)

	def _validatePerform(self, check_only, stepSetAndCategory):
		logger = Logger()
		ctx = Context(logger)
		stepExecutor = StepExecutor(ctx, logger)
		ctx.check_only = check_only
		return self._validateVerb(stepExecutor, stepSetAndCategory, "perform", stepExecutor.perform)

	def _validateAllSteps(self, check_only, allStepSetAndCategory, performMethodDescription, performMethod):
		allPassed = True
		for stepSetAndCategory in allStepSetAndCategory:
			category = stepSetAndCategory[1]
			categoryName = StepResultCategory.Descriptions[category]
			if not performMethod(check_only,
				stepSetAndCategory):
				allPassed = False
				print("Category {0} failed {1} validation when check_only = {2}".format(categoryName, performMethodDescription, check_only))
		if not allPassed:
			print("Not all steps passed {0} validation when check_only = {1}".format(performMethodDescription, check_only))
		return allPassed

	def _validateAllStepsAndStates(self, checkOnlyAndStepSets, performMethodDescription, performMethod):
		allPassed = True
		for state in checkOnlyAndStepSets:
			check_only = state[0]
			allStepSetAndCategory = state[1]
			allPassed = allPassed and self._validateAllSteps(check_only, allStepSetAndCategory, performMethodDescription, performMethod)
		return allPassed

	def testResultValues(self):
		'''Test that all Results have a description and value.
		'''
		self.assertFalse(StepResultReason.Count < len(StepResultReason.Descriptions),
			"A result reason is missing a description")
		self.assertFalse(StepResultReason.Count < len(StepResultReason.ToCategory),
			"A result reason is missing a category")
		self.assertFalse(StepResultReason.Count > len(StepResultReason.Descriptions),
			"There are too many result descriptions")
		self.assertFalse(StepResultReason.Count > len(StepResultReason.ToCategory),
			"There are too many mappings from a result to a category")

	def testFromBool(self):
		'''Test that StepResult.fromBool()
		returns expected values.
		'''

		self.assertTrue(StepResult.fromBool(True) == StepResult.Success)
		self.assertTrue(StepResult.fromBool(False) == StepResult.Fail)

	def testPerform(self):
		'''Test that StepExecutor.perform()
		returns expected values from steps and does not
		raise exceptions during execution.
		'''
		checkOnlyStates = ((True, AllCheckSteps), (False, AllSteps))
		self.assertTrue(self._validateAllStepsAndStates(checkOnlyStates, "perform", self._validatePerform))

	def testCheck(self):
		'''Test that StepExecutor.check()
		returns expected values from steps and does not
		raise exceptions during execution.
		'''
		checkOnlyStates = ((True, AllCheckSteps), (False, AllCheckSteps))
		self.assertTrue(self._validateAllStepsAndStates(checkOnlyStates, "check", self._validateCheck))


def suite():
	'''Returns all cases that should be run
	to test this part of the application.
	'''

	all_cases = (ShellManagerTest, PackageManagerTest, StepExecutorTest)
	all_suites = []

	for case in all_cases:
		all_suites.append(unittest.TestLoader().loadTestsFromTestCase(case))

	return unittest.TestSuite(all_suites)

if __name__ == "__main__":
	unittest.TextTestRunner(verbosity=2).run(suite())



'''Unit tests for version module.
'''

import unittest
import operator
from distutils.version import LooseVersion
from ...setup import Version, VersionPair

class VersionTest(unittest.TestCase):
	'''Unit tests for version module.
	'''

	def testVersionReprWorks(self):
		'''Tests that version representations work.
		'''

		exampleVersionStr = "1.0"
		exampleVersion = Version("1.0")
		print("Repr of Version: '{0}'".format(exampleVersion))
		stringIsStored = exampleVersion.string == exampleVersionStr
		if not stringIsStored:
			failStr = "Failed: expected version to have string '{0}', got '{1}' instead"
			print(failStr.format(exampleVersionStr, exampleVersion.string))
		versionIsSet = exampleVersion.value == LooseVersion(exampleVersionStr)
		if not versionIsSet:
			print("Failed: version does not match its LooseVersion equivalent")

		self.assertTrue(stringIsStored and versionIsSet)

	def testVersionComparsionWorks(self):
		'''Tests that version comparisons work.
		'''

		version1 = (Version("1.0"), 1)
		version2 = (Version("1.1"), 2)
		version3 = (Version("1.1.1"), 3)
		versions = (version1, version2, version3)
		comparisonsWork = True
		for pair1 in versions:
			for pair2 in versions:
				#Do all comparisons.
				ops = [operator.eq, operator.lt, operator.le, operator.gt, operator.ge]
				for op in ops:
					expectedResult = op(pair1[1], pair2[1])
					if op(pair1[0], pair2[0]) != expectedResult:
						print("Version.{0}() doesn't return {1} as expected".format(op.__name__, expectedResult))
						comparisonsWork = False

		self.assertTrue(comparisonsWork, "Some comparisons don't evaluate as expected")

	def testVersion(self):
		'''Tests what happens when an invalid version is created.
		'''

		print("Creating invalid version...")
		badVersionStr = "roijs"
		try:
			print("Repr of Version('{0}''): '{1}'".format(badVersionStr, Version(badVersionStr)))
		except Exception as e:
			print("Trying to create an invalid version returns exception {0}".format(e))

	def testVersionPairRepr(self):
		'''Tests version pair representation.
		'''
		print("")
		print("Repr of VersionPair: '{0}'".format(VersionPair("1.0", "2.0")))

	def testVersionLimits(self):
		'''Test that omitting either or both
		of the limit strings makes that limit None.
		'''
		testLimits = (("1.0", None), (None, "1.0"), ("1.0", "2.0"))
		limitsCanBeNone = True
		for limitPair in testLimits:
			expectedMin = limitPair[0]
			if expectedMin:
					expectedMin = Version(expectedMin)
			expectedMax = limitPair[1]
			if expectedMax:
					expectedMax = Version(expectedMax)
			currVersion = VersionPair(limitPair[0], limitPair[1])
			actualMin = currVersion.minimum
			actualMax = currVersion.maximum
			minInvalid = expectedMin is not None and actualMin != expectedMin
			maxInvalid = expectedMax is not None and actualMax != expectedMax
			if minInvalid or maxInvalid:
				reason = ""
				if minInvalid and maxInvalid:
						errString = ("(minimum and maximum invalid, should be min = {0},"
						" max = {1}, is min = {2}, max = {3})")
						reason = errString.format(expectedMin, expectedMax, actualMin, actualMax)
				elif minInvalid:
						reason = "(minimum invalid, should be {0}, is {1})".format(expectedMin, actualMin)
				elif maxInvalid:
						reason = "(maximum invalid, should be {0}, is {1})".format(expectedMax, actualMax)
				errString = "Failure: test limits {0} generated version pair {1} {2}"
				print(errString.format(limitPair, currVersion, reason))
				limitsCanBeNone = False

		self.assertTrue(limitsCanBeNone)

	def testMinCanEqualMax(self):
		'''Test that min == max doesn't raise exceptions.
		'''
		currVersion = VersionPair("1.0", "1.0")
		self.assertTrue(currVersion is not None)

	def testInvalidRangesGenerateExceptions(self):
		'''Test that min > max does raise exceptions.
		'''
		try:
			currVersion = VersionPair("1.1.0", "1.0")
			errString = "Failure: invalid version pair {0} isn't considered invalid"
			self.assertTrue(False, errString.format(currVersion))
		except RuntimeError:
			pass

	def testVersionIsAcceptable(self):
		'''Test VersionPair.versionIsAcceptable().
		'''

		versionRange = VersionPair("2.0", "3.0")
		acceptableVersion = Version("2.0")
		oldVersion = Version("1.0")
		newVersion = Version("4.0")

		#	* Too old versions return False.
		tooOldWorks = not versionRange.versionIsAcceptable(oldVersion)
		if not tooOldWorks:
			print("Invalid version {0} accepted by version range {1}".format(oldVersion,
			versionRange))

		#	* Too new versions return False.
		tooNewWorks = not versionRange.versionIsAcceptable(newVersion)
		if not tooNewWorks:
			print("Invalid version {0} accepted by version range {1}".format(newVersion,
			versionRange))

		#	* Everything else returns True.
		inRangeWorks = versionRange.versionIsAcceptable(acceptableVersion)
		if not inRangeWorks:
			print("Valid version {0} not accepted by version range {1}".format(acceptableVersion,
			versionRange))

		self.assertTrue(tooOldWorks and
			tooNewWorks and
			inRangeWorks)

if __name__ == "__main__":
	suite = unittest.TestLoader().loadTestsFromTestCase(VersionTest)
	unittest.TextTestRunner(verbosity=2).run(suite)



'''Unit tests for registry module.
'''

import unittest
import platform
from ...setup import Registry

if platform.system() == "Windows":
	try:
		import winreg as _winreg
	except ImportError:
		import _winreg

	class RegistryTest(unittest.TestCase):
		'''Unit tests for registry module.
		'''

		_KeyShouldExist = "SOFTWARE\\"
		_KeyShouldNotExist = "RFUIOUBSEF\\"
		_SubkeyShouldExist = ""

		def testOpenPath(self):
			'''Test Registry.openPath().
			'''

			#Test: valid keys return a key object (not None).
			hKey = _winreg.HKEY_LOCAL_MACHINE
			validWorks = Registry.openPath(hKey, self.__class__._KeyShouldExist) is not None
			if not validWorks:
				print("Valid key {0}{1} could not be opened".format(hKey, self.__class__._KeyShouldExist))
			#Test: invalid keys return None.
			invalidWorks = Registry.openPath(hKey, self.__class__._KeyShouldNotExist) is None
			if not invalidWorks:
				print("Invalid key {0}{1} was somehow opened".format(hKey, self.__class__._KeyShouldNotExist))

			return validWorks and invalidWorks

		#Can't really test this unless there's definitely
		#a MSVC installation on the test system,
		#so not implementing this.
		#def testCheckKeyExists(self):
		#	#	checkKeyExists():
		#	#		* Valid subkeys return True.
		#	#		* Invalid subkeys return False.
		#	raise NotImplementedError()
else:
	#Registry calls will fail/return None if not under Windows,
	#so skip tests in this situation.
	class RegistryTest(unittest.TestCase):
		'''Dummy unit test when running on non-Windows platform.
		'''
		def testRegistry(self):
			'''Dummy test informing tester that this suite is being skipped.
			'''
			print("")
			print("RegistryTest requires Windows, skipping")

if __name__ == "__main__":
	suite = unittest.TestLoader().loadTestsFromTestCase(RegistryTest)
	unittest.TextTestRunner(verbosity=2).run(suite)



'''Setup file for gittools package.
'''
from setuptools import setup, find_packages

setup(
	# Application name:
	name="jtitor_basis",
	# Version number (initial):
	version="0.1.4",
	# Application author details:
	author="jTitor",
	author_email="name@addr.ess",
	# Packages
	packages=find_packages(),
	# Include additional files into the package
	include_package_data=True,
	# Details
	url="http://pypi.python.org/pypi/jtitor_basis/0.1.4/",
	license="MIT",
	description=("Framework for cross-platform scripts that run in stages. "
	"Privately listed since it's not quite done."),
	#long_description=open("readme.md").read(),
	# Dependent packages (distributions)
	install_requires=[
		"colorama",
	],
	entry_points={
		'console_scripts': ['testbasis=basis.testbasis:main']
	}
)



'''Runs all unit tests.
'''

import unittest
from .arguments import ArgumentsTest
from .context import ContextTest
from .log import LogTest
from .registry import RegistryTest
from .step.stepbase import StepExecutorTest
from .step.shellmanager import ShellManagerTest
from .step.packagemanager import PackageManagerTest
from .version import VersionTest

class AllUnitTest:
	'''Runs all unit tests.
	'''

	@classmethod
	def suite(cls):
		'''Returns all unit tests.
		'''

		allCases = (ArgumentsTest,
			ContextTest,
			LogTest,
			RegistryTest,
			StepExecutorTest,
			ShellManagerTest,
			PackageManagerTest,
			VersionTest)
		allSuites = []

		for case in allCases:
			allSuites.append(unittest.TestLoader().loadTestsFromTestCase(case))

		return unittest.TestSuite(allSuites)

if __name__ == "__main__":
	unittest.TextTestRunner(verbosity=2).run(AllUnitTest.suite())



'''State data passed between different elements of the program.
'''

import sys
import os
import platform
import ctypes
from .logging import Logger, LogLevel
from .arguments import Arguments

#Enums.
class DistroCode(object):
	'''Enumeration for Linux distributions.
	'''

	Debian = 0
	RedHat = 1
	Count = 2
	Descriptions = ["Debian-based", "Red Hat-based", "N/A"]

class OSCode(object):
	'''Enumeration for operating systems.
	'''

	Windows = 0
	Linux = 1
	MacOS = 2
	Other = 3
	Count = 4
	AllCodes = (Windows, Linux, MacOS, Other)
	Descriptions = ["Windows", "Linux", "macOS", "Unknown", "N/A"]

DefaultEnvironmentVariableFile = "~/.bashrc"
#Technically need Windows 7 for setx.

class Context(object):
	'''State data passed between different elements of the program.
	These shouldn't really be modified after creation.
	'''

	def isValid(self):
		'''Returns True if this context can be used
		by a StepExecutor, False otherwise.
		'''
		requiredFields = ((self.packageRunInstallerTool, "packageRunInstallerTool"),
			(self.escalateTool, "escalateTool"),
			(self.newShellTool, "newShellTool"),
			(self.downloadTool, "downloadTool"),
			(self.pythonVersion, "pythonVersion")
			)
		#A valid context has all required fields
		#set.
		allFieldsValid = True
		for field in requiredFields:
			#Report all fields that aren't set.
			if not field[0]:
				self._log.error("Context.{0} not set!".format(field[1]))
				allFieldsValid = False
		return allFieldsValid

	def _makeFields(self):
		self.osIdentifier = OSCode.Count
		self.osDistribution = DistroCode.Count
		#The bootstrapper works on this
		#operating system.
		self.osSupported = False
		self.packageInstallTool = ""
		self.packageUpdateTool = ""
		self.packageCheckInstalledTool = ""
		#This is a format string!!!
		self.packageRunInstallerTool = ""
		#Command used to escalate privileges.
		#Is a format string.
		self.escalateTool = ""
		self.userIsAdmin = False
		#Command used to open a new shell.
		self.newShellTool = ""
		#Command used to download files remotely.
		#Is a format string:
		#	0: The source URL.
		#	1: The destination URL.
		self.downloadTool = ""
		#List of steps that were run but failed.
		#TODO: Can be moved to step executor/Basis?
		self.failedSteps = []
		#If true, performStep() calls only perform the check.
		#Used for testing purposes.
		self.check_only = False
		#If true, print out low-priority info too.
		self.verbose = False
		#If true, print out debugging info.
		self.debug = False
		#If true, warnings that'd ordinarily throw
		#will not throw.
		self.permit_warnings = False
		#If true, perform a check of all steps after an install run.
		self.verify = False
		self.pythonVersion = None
		#The file to read/write environment variables
		#to. Does nothing on Windows.
		self.environmentVariableFile = ""
		#Boolean flags used by implementing program.
		self.extraFlags = ()
		#The parsed arguments as an argparse object.
		self.parsedArguments = None
		self._log = None

	def _osCodeFromIdentifier(self):
		identifier = platform.system()
		result = OSCode.Count
		if identifier == "Windows":
			result = OSCode.Windows
		elif identifier == "Linux":
			result = OSCode.Linux
		elif identifier == "Darwin":
			result = OSCode.MacOS
		else:
			result = OSCode.Other
		return result

	def _initFields(self, logger, args=None, extraFlags=()):
		#Parse flags.
		self._log = logger
		self.extraFlags = extraFlags
		self.parsedArguments = Arguments.parse(args, extraFlags)
		self.check_only = self.parsedArguments.check_only
		self.permit_warnings = self.parsedArguments.permit_warnings
		self.osIdentifier = self._osCodeFromIdentifier()
		self.verbose = self.parsedArguments.verbose or self.parsedArguments.debug
		self.debug = self.parsedArguments.debug
		#Set the logger's print level.
		if self.debug:
			self._log._minimumPrintableLevel = LogLevel.Debug
		elif self.verbose:
			self._log._minimumPrintableLevel = LogLevel.Verbose

		#If --check-only is set, this should be false.
		self.verify = (not self.check_only) and self.parsedArguments.verify

		self.pythonVersion = sys.version_info

		#Do platform-dependent calls.
		if self.osIsWindows():
			self.osSupported = True
			#/i = install, /qn = no UI.
			self.packageRunInstallerTool = "msiexec /i \\\"{0}\\\" /qn /norestart"
			self.escalateTool = "runas /user:Administrator \\\"{0}\\\""
			self.userIsAdmin = ctypes.windll.shell32.IsUserAnAdmin()
			self.newShellTool = "cmd"
			self.downloadTool = "(New-Object System.Net.WebClient).DownloadFile(\\\"{0}\\\", \\\"{1}\\\")"
		elif self.osIsLinux() or self.osIsMacOS():
			self.osSupported = True
			self.environmentVariableFile = DefaultEnvironmentVariableFile
			self.escalateTool = "sudo {0}"
			self.userIsAdmin = os.getuid() == 0
			self.newShellTool = "bash"
			self.downloadTool = "curl {0} -o {1}"
			if self.osIsLinux():
				#Check distribution to see what
				#package manager to use.
				distroName = platform.linux_distribution()[0]
				#Debians use apt.
				if distroName in ("debian", "ubuntu"):
					self.osDistribution = DistroCode.Debian
					self.packageInstallTool = "apt-get -y install"
					self.packageUpdateTool = "apt-get -y update && apt-get -y dist-upgrade"
					self.packageCheckInstalledTool = "dpkg -s"
					self.packageRunInstallerTool = "dpkg -i {0} && apt-get install -y -f"
				#Fedoras use yum.
				elif distroName in ("redhat", "centos", "CentOS Linux"):
					#TODO: Technically depends on version - Red Hat 22+ uses DNF.
					self.osDistribution = DistroCode.RedHat
					self.packageInstallTool = "yum -y install"
					self.packageUpdateTool = "yum -y update"
					self.packageCheckInstalledTool = "rpm -q"
					self.packageRunInstallerTool = "yum install -y {0}"
				#Otherwise the OS isn't supported.
				else:
					#Report the problem and abort.
					self.osSupported = False
					self._log.error("Distribution " + distroName + " isn't supported!")
			elif self.osIsMacOS():
				self.packageInstallTool = "brew install"
				self.packageUpdateTool = "brew update && brew upgrade && brew cleanup"
				self.packageCheckInstalledTool = "brew ls --versions"
				self.packageRunInstallerTool = "sudo installer -verboseR -pkg \"{0}\" -target /"

	def __init__(self, logger, args=None, extraFlags=()):
		assert isinstance(logger, Logger)
		self._makeFields()
		self._initFields(logger, args, extraFlags)

	def osIsWindows(self):
		'''Returns true if the current platform
		is a version of Windows.
		'''
		return self.osIdentifier == OSCode.Windows

	def osIsLinux(self):
		'''Returns true if the current platform
		is a version of Linux.
		'''
		return self.osIdentifier == OSCode.Linux

	def osIsMacOS(self):
		'''Returns true if the current platform
		is a version of macOS.
		'''
		return self.osIdentifier == OSCode.MacOS

	def printContextDescription(self):
		'''Prints a full description of the
		context's state.
		'''
		self._log.info("Current context:")
		self._log.info("\tIdentifier: '{0}'".format(OSCode.Descriptions[self.osIdentifier]))
		logMethod = self._log.success
		if not self.osSupported:
			logMethod = self._log.error
		logMethod("\tOS supported by basis script: {0}".format(self.osSupported))
		self._log.info("\tCheck-only mode: {0}".format(self.check_only))
		self._log.info("\tVerbose mode: {0}".format(self.verbose))
		self._log.info("\tDebug mode: {0}".format(self.debug))
		self._log.info("\tDefault package install invocation: '{0}'".format(self.packageInstallTool))
		self._log.info("\tDefault package update invocation: '{0}'".format(self.packageUpdateTool))
		self._log.info("\tDefault package verify installed invocation: \
		'{0}'".format(self.packageCheckInstalledTool))
		self._log.info("\tDefault package run installer invocation: \
		'{0}'".format(self.packageRunInstallerTool))
		self._log.info("\tDefault escalation invocation: '{0}'".format(self.escalateTool))
		self._log.info("\tDefault shell invocation: '{0}'".format(self.newShellTool))
		self._log.info("\tDefault file download invocation: '{0}'".format(self.downloadTool))

		#Print platform-specific context data.
		#Print all fields even if they'd be invalid on this platform.
		self._log.info("\tPOSIX-only context:")
		self._log.info("\t\tEnvironment variable file: '{0}'".format(self.environmentVariableFile))
		self._log.info("\tLinux-only context:")
		self._log.info("\t\tDistribution number: {0} ({1})".format(self.osDistribution,
		DistroCode.Descriptions[self.osDistribution]))

	def printContextSummary(self):
		'''Prints a short summary of the platform
		the context was derived from.
		'''
		distroString = ""
		if self.osIsLinux():
			distroString = ", distribution: {0}".format(DistroCode.Descriptions[self.osDistribution])
		self._log.info("Detected OS: {0}{1}".format(self.osIdentifier, distroString))
